# flake8: noqa: F401
r"""
This file is in the process of migration to `torch/ao/quantization`, and
is kept here for compatibility while the migration process is ongoing.
If you are adding a new entry/functionality, please, add it to the
appropriate files under `torch/ao/quantization/fx/`, while adding an import statement
here.
"""
from torch.ao.quantization.fx.graph_module import (
    GraphModule,
    FusedGraphModule,
    ObservedGraphModule,
    _is_observed_module,
    ObservedStandaloneGraphModule,
    _is_observed_standalone_module,
    QuantizedGraphModule
)
