#pragma once

// @generated by torchgen/gen.py from Function.h

#include <ATen/Context.h>
#include <ATen/DeviceGuard.h>
#include <ATen/TensorUtils.h>
#include <ATen/TracerMode.h>
#include <ATen/core/Generator.h>
#include <ATen/core/Reduction.h>
#include <ATen/core/Tensor.h>
#include <c10/core/Scalar.h>
#include <c10/core/Storage.h>
#include <c10/core/TensorOptions.h>
#include <c10/util/Deprecated.h>
#include <c10/util/Optional.h>



#include <ATen/ops/_chunk_grad_outputs_efficient_attention_ops.h>

namespace at {


// aten::_chunk_grad_outputs_efficient_attention(Tensor query, Tensor key, Tensor value, bool is_causal=False) -> bool
inline bool _chunk_grad_outputs_efficient_attention(const at::Tensor & query, const at::Tensor & key, const at::Tensor & value, bool is_causal=false) {
    return at::_ops::_chunk_grad_outputs_efficient_attention::call(query, key, value, is_causal);
}

}
